/***************************************************************************
 *
 *                              FREEZING DROID
 *    This is a kernel module to visualize the bit error rate arising from
 *              cold boot attacks on Android-based smartphones
 *
 * Copyright (C) 2012 Tilo Mueller <tilo.mueller@informatik.uni-erlangen.de>
 *
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms and conditions of the GNU General Public License,
 * version 2, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place - Suite 330, Boston, MA 02111-1307 USA.
 *
 ***************************************************************************/

#include <linux/highmem.h>
#include <linux/device.h>
#include <linux/module.h>
#include <linux/ioport.h>
#include <linux/kernel.h>
#include <linux/types.h>
#include <linux/pfn.h>
#include "droid.h"

#define PAGE_ADDRESS (void *)0xc5000000		/* phys address of the page we test */
int check = 0;					/* 1 = check page  or  0 = set page */
int droid = 1;					/* 1 = fill droid  or  0 = fill FFs */
module_param(check, int, S_IRUGO);
module_param(droid, int, S_IRUGO);
extern struct resource iomem_resource;

/* number of bits that differ between byte a and b */
static inline int bitdiff(uint8_t a, uint8_t b)
{
        int i, count = 0;
	uint8_t d = a ^ b;
        for (i = 0; i < 8; i++)
		count += (d>>i) & 1;
        return count;
}

/* this module must be run twice: before and after a cold boot attack.
 * before cold boot, run it with "insmod freezing_droid.ko" to fill a page.
 * after cold boot, run it with "insmod freezing_droid.ko check=1" to analyze a page.
 */
int init_module(void)
{
	int j, bit_errors = 0;
	struct resource *res;
	for (res = iomem_resource.child; res ; res = res->sibling) {
		if (strncmp(res->name, "System RAM", sizeof("System RAM")))
			continue;
		else {
			mm_segment_t fs;
			resource_size_t i;
			struct page *p;
			uint8_t *v;
			fs = get_fs();
			set_fs(KERNEL_DS);
			for (i = res->start; i < res->end; i += PAGE_SIZE) { /* search for desired page at PAGE_ADDRESS */
				p = pfn_to_page(PFN_DOWN(i));
				v = (uint8_t *)kmap(p);
				if (v == PAGE_ADDRESS) {
					if (check) { /* analyze recently filled page */	
						if (droid) { /* visualize as bitmap */
							printk("[FREEZING_DROID] unsigned char droid_bmp[] = {");
							for (j=0; j<sizeof(droid_bmp); j++)
								printk("0x%02x,",v[j]);
							printk("}\n");
						}
						else { /* or calculate bit error rate */
							for (j=0; j<PAGE_SIZE; j++) {
								if (v[j] != 0xFF) {
									bit_errors += bitdiff(v[j],0xFF);
								}
							}
							printk("[FREEZING_DROID] Total bit errors: %i\n",bit_errors);
							printk("[FREEZING_DROID] Bit error rate: %li\n",(bit_errors*100)/(PAGE_SIZE*8));
						}
					}
					else { /* fill a page .. */
						if (droid) { /* .. with a droid bitmap */
							for (j=0; j<sizeof(droid_bmp); j++)
								v[j] = droid_bmp[j];
						}
						else { /* .. or with 0xFF bytes */
							for (j=0; j<PAGE_SIZE; j++)
								v[j] = 0xFF;
						}
						printk("[FREEZING_DROID] Page filled with %li bytes.\n",PAGE_SIZE);
					}
				}
				kunmap(p);
			}
			set_fs(fs);
		}
	}
	return 0;
}

void cleanup_module(void) { }
MODULE_LICENSE("GPL");
MODULE_AUTHOR ("Tilo Mueller, tilo.mueller@informatik.uni-erlangen.de");
MODULE_DESCRIPTION ("Analysis of the bit error rate from cold boot attacks.");
